//
// PODSteam - Loader for Old CS Bots
// Copyright (c) 2005, by Wei Mingzhi
// Copyright (c) 2006, by Dmitry Zhukov
//
// BotAim - Human-like Bot Aiming Plugin 
// Copyright (c) 2004, by Pierre-Marie Baty
//
// Note: Using Botaim 1.5, because version 2 doesn't work with POD-Bot
//

#include "podsteam.h"

cvar_t pstm_turn_method = {"pstm_turn_method", "2", FCVAR_EXTDLL};
cvar_t pstm_turn_skill = {"pstm_turn_skill", "4", FCVAR_EXTDLL};
cvar_t pstm_aim_fix = {"pstm_aim_fix", "1", FCVAR_EXTDLL};

void CBotAim::RegisterCVars (void)
{
   // ask the engine to register the CVARs this plugin uses
   CVAR_REGISTER (&pstm_turn_method);
   CVAR_REGISTER (&pstm_aim_fix);
   CVAR_REGISTER (&pstm_turn_skill);
}

void CBotAim::OnFramePost (void)
{
   int player_index;

   // get the turn method and do a fool-proof check
   iTurnMethod = CVAR_GET_FLOAT ("pstm_turn_method");
   
   if (iTurnMethod < 0) 
   {
      CVAR_SET_FLOAT ("pstm_turn_method", 0); 
      iTurnMethod = 0; 
   }
   else if (iTurnMethod > 2) 
   { 
      CVAR_SET_FLOAT ("pstm_turn_method", 2); 
      iTurnMethod = 2; 
   }

   // get the turn skill and do a fool-proof check
   m_iTurnSkill = CVAR_GET_FLOAT ("pstm_turn_skill");
   if (m_iTurnSkill < 1) 
   { 
      CVAR_SET_FLOAT ("pstm_turn_skill", 1); 
      m_iTurnSkill = 1; 
   }

   // get whether we need to fix "aimbug fix" bug or not
   m_bAimFix = (CVAR_GET_FLOAT ("pstm_aim_fix") > 0);

   // cycle through all players
   for (player_index = 1; player_index <= gpGlobals->maxClients; player_index++)
   {
      edict_t *pPlayer = INDEXENT (player_index);

      if (FNullEnt (pPlayer) || pPlayer->free)
         continue; // skip invalid players

      // if this player is a bot AND user specified a custom aim method, make it aim and turn
      if ((pPlayer->v.flags & FL_FAKECLIENT) && (iTurnMethod != AIM_GENUINE) && !IsYaPB (pPlayer))
         PointGun (pPlayer, TRUE); // update and save this bot's view angles

      // systematically wrap all player angles to avoid engine freezes
      pPlayer->v.angles = WrapAngles (pPlayer->v.angles);
      pPlayer->v.v_angle = WrapAngles (pPlayer->v.v_angle);
      pPlayer->v.idealpitch = WrapAngle (pPlayer->v.idealpitch);
      pPlayer->v.ideal_yaw = WrapAngle (pPlayer->v.ideal_yaw);
   }
}

bool CBotAim::OnTraceLine (const float *v1, const float *v2, int fNoMonsters, edict_t *pentToSkip, TraceResult *ptr)
{
   // check to see if a bot is sending this traceline AND it is firing AND we are taking over the bot aiming code
   if (!FNullEnt (pentToSkip) && (pentToSkip->v.flags & FL_FAKECLIENT) && !IsYaPB (pentToSkip) && (pentToSkip->v.button & (IN_ATTACK | IN_ATTACK2)) && (iTurnMethod != AIM_GENUINE))
   {
      PointGun (pentToSkip, FALSE); // update the bot's view angles in advance, but don't save them
      MAKE_VECTORS (pentToSkip->v.v_angle); // build base vectors in bot's actual view direction

      // trace the line considering the ACTUAL view angles of the bot, not the ones it claims to have
      TRACE_LINE (pentToSkip->v.origin + pentToSkip->v.view_ofs,
                  pentToSkip->v.origin + pentToSkip->v.view_ofs + gpGlobals->v_forward * 9999,
                  fNoMonsters, pentToSkip, ptr);

      return false; // and don't let the engine be fooled by the bot's lies, goddamnit! :)
   }
   return true;
}

void CBotAim::PointGun (edict_t *pEdict, bool save_angles)
{
   // this function is called every frame for every bot. Its purpose is to make the bot
   // move its crosshair to the direction where it wants to look. There is some kind of
   // filtering for the view, to make it human-like.

   float speed; // speed : 0.1 - 1
   Vector v_deviation;

   v_deviation = WrapAngles (Vector (pEdict->v.idealpitch, pEdict->v.ideal_yaw, 0) - m_vViewAngle[ENTINDEX (pEdict) - 1]);

   // do we want old RACC style aiming ?
   if (iTurnMethod == AIM_RACC_OLD)
   {
      // limit the turn skill
      speed = (m_iTurnSkill * 3 < 1 / gpGlobals->frametime ? m_iTurnSkill * 3 : 1 / gpGlobals->frametime);

      // offset yaw and pitch at some percentage of the deviation, based on skill and frame time.
      pEdict->v.yaw_speed = v_deviation.y * gpGlobals->frametime * speed;
      pEdict->v.pitch_speed = v_deviation.x * gpGlobals->frametime * speed;
   }

   // else do we want new RACC style aiming ?
   else if (iTurnMethod == AIM_RACC)
   {
      // if bot is aiming at something, aim fast, else take our time...
      if (IsAim (pEdict))
         speed = 0.7 + (m_iTurnSkill - 1) / 10; // fast aim
      else
         speed = 0.2 + (m_iTurnSkill - 1) / 20; // slow aim

      // thanks Tobias "Killaruna" Heimann and Johannes "@$3.1415rin" Lampel for this one
      pEdict->v.yaw_speed = (pEdict->v.yaw_speed * exp (log (speed / 2) * gpGlobals->frametime * 20)
                             + speed * v_deviation.y * (1 - exp (log (speed / 2) * gpGlobals->frametime * 20)))
                            * gpGlobals->frametime * 20;
      pEdict->v.pitch_speed = (pEdict->v.pitch_speed * exp (log (speed / 2) * gpGlobals->frametime * 20)
                               + speed * v_deviation.x * (1 - exp (log (speed / 2) * gpGlobals->frametime * 20)))
                              * gpGlobals->frametime * 20;

      // influence of y movement on x axis, based on skill (less influence than x on y since it's
      // easier and more natural for the bot to "move its mouse" horizontally than vertically)
      if (pEdict->v.pitch_speed > 0)
         pEdict->v.pitch_speed += pEdict->v.yaw_speed / (1.5 * (1 + m_iTurnSkill));
      else
         pEdict->v.pitch_speed -= pEdict->v.yaw_speed / (1.5 * (1 + m_iTurnSkill));

      // influence of x movement on y axis, based on skill
      if (pEdict->v.yaw_speed > 0)
         pEdict->v.yaw_speed += pEdict->v.pitch_speed / (1 + m_iTurnSkill);
      else
         pEdict->v.yaw_speed -= pEdict->v.pitch_speed / (1 + m_iTurnSkill);
   }

   // move the aim cursor
   pEdict->v.v_angle = WrapAngles (m_vViewAngle[ENTINDEX (pEdict) - 1] + Vector (pEdict->v.pitch_speed, pEdict->v.yaw_speed, 0)); 

   // set the body angles to point the gun correctly
   pEdict->v.angles.x = WrapAngle (pEdict->v.v_angle.x / 3);
   pEdict->v.angles.y = WrapAngle (pEdict->v.v_angle.y);
   pEdict->v.angles.z = 0;

   // did the bot maker not get rid of Paulo-La-Frite's "bugfix" bug ?
   if (m_bAimFix)
      pEdict->v.angles.x = WrapAngle (-pEdict->v.angles.x); // if the bug is still here, fix it again

   // if we want to save the bot's view angles, remember these view angles
   if (save_angles)
      m_vViewAngle[ENTINDEX (pEdict) - 1] = WrapAngles (pEdict->v.v_angle);

   return;
}

bool CBotAim::IsAim (edict_t *pEdict)
{
   TraceResult tr;

   if (FNullEnt (pEdict))
      return FALSE; // reliability check

   // first make vectors in the direction pEdict wants to look
   MAKE_VECTORS (WrapAngles (Vector (pEdict->v.idealpitch, pEdict->v.ideal_yaw, 0)));
   TRACE_LINE (pEdict->v.origin + pEdict->v.view_ofs,
               pEdict->v.origin + pEdict->v.view_ofs + gpGlobals->v_forward * 9999,
               dont_ignore_monsters, pEdict, &tr);
   if (strcmp (STRING (tr.pHit->v.classname), "player") == 0)
      return (TRUE); // bot is aiming at a player

   return (FALSE); // bot is aiming at anything but a player
}

Vector VecToAngles (const Vector &v_vector)
{
   float angles[3];

   VEC_TO_ANGLES (v_vector, angles);

   return (WrapAngles (Vector (angles)));
}


float AngleOfVectors (Vector v1, Vector v2)
{
   if ((v1.Length () == 0) || (v2.Length () == 0))
      return (0); // avoid zero divide

   return (WrapAngle (acos (DotProduct (v1, v2) / (v1.Length () * v2.Length ())) * 180 / M_PI));
}


float WrapAngle (float angle)
{
   // this function adds or substracts 360 enough times needed to the given angle in
   // order to set it into the range [-180, 180) and returns the resulting angle. Letting
   // the engine have a hand on angles that are outside these bounds may cause the game
   // to freeze by screwing up the engine math code.

   return (360.0 / 65536) * ((int) ((angle + 180) * (65536.0 / 360.0)) & 65535) - 180;
}


Vector WrapAngles (Vector vAngles)
{
   vAngles.x = WrapAngle (vAngles.x);
   vAngles.y = WrapAngle (vAngles.y);
   vAngles.z = 0;

   return vAngles;
}