//
// PODSteam - Loader for Old CS Bots
// Copyright (c) 2005, by Wei Mingzhi
// Copyright (c) 2006, by Dmitry Zhukov
//
// Linkent code is courtesy of Jussi Kivilinna

#include "podsteam.h"

#ifdef PF_WIN32
   
IMAGE_NT_HEADERS *GetNTHeader (dynlib_t hModule)
{
   // this function reading nt headers from valid windows portable executable, and then returning it
      
	union 
   { 
		uint32 ulMem;
		IMAGE_DOS_HEADER *pDosHdr;
		IMAGE_NT_HEADERS *pPeHdr;
	} kMem;
	
	kMem.ulMem = (uint32) hModule;
   
   if(IsBadReadPtr (kMem.pDosHdr, sizeof (*kMem.pDosHdr)) || (kMem.pDosHdr->e_magic != IMAGE_DOS_SIGNATURE))
		return FALSE;
	
	kMem.ulMem = RVA_TO_VA (hModule, kMem.pDosHdr->e_lfanew);
	
   if(IsBadReadPtr(kMem.pPeHdr, sizeof(*kMem.pPeHdr)) || (kMem.pPeHdr->Signature != IMAGE_NT_SIGNATURE))
		return FALSE;
	
	return (kMem.pPeHdr);
}


IMAGE_EXPORT_DIRECTORY *GetExportTable (dynlib_t hModule)
{
	union 
   { 
		uint32 ulMem;
		void *pVoid;
		IMAGE_DOS_HEADER *pDosHdr;
		IMAGE_NT_HEADERS *pPeHdr;
		IMAGE_EXPORT_DIRECTORY *pExpDir;
	} kMem;
	
	kMem.pPeHdr = GetNTHeader (hModule);
	if(!kMem.pPeHdr)
		return FALSE;
	
	if(!kMem.pPeHdr->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress)
		return FALSE;
	
	kMem.ulMem = RVA_TO_VA (hModule, kMem.pPeHdr->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress);
	
   if(IsBadReadPtr (kMem.pExpDir, sizeof (*kMem.pExpDir)))
		return FALSE;
	
	return(kMem.pExpDir);
}


int SortOrdinalList (const sortnames_t *pA, sortnames_t *pB)
{
   // this function sorts GAME library exports
      
	const char *cszA = reinterpret_cast <const char *> (pA->ulName);
	const char *cszB = reinterpret_cast <const char *> (pB->ulName);
	
	return (strcmp (cszA, cszB));
}


int CombineExportTables (void)
{
   // this function combines our & gamedll exports, for bind all game dll exported function ot ours
      
	IMAGE_EXPORT_DIRECTORY *pSelfExport;
	IMAGE_EXPORT_DIRECTORY *pGameExport;
	
	uint32 ulNumberOfFunctions;
	uint32 ulNumberOfNames;
	uint32 *pulFunctions;
	uint32 *pulNames;
	uint16 *pulNameOrdinals;
	sortnames_t *pSort;
	
	uint32 i, u;
	uint32 ulFunctionCount;
	uint32 ulNameCount;
	uint32 ulListFix;

   assert (g_hSelfLib && g_hGameLib);
	
	pSelfExport = GetExportTable (g_hSelfLib);
	pGameExport = GetExportTable (g_hGameLib);

	if((pSelfExport == NULL) || (pGameExport == NULL))
      other_gFunctionTable.pfnSys_Error ("Couldn't initialize dynamic linkents.");

	ulNumberOfFunctions = pSelfExport->NumberOfFunctions + pGameExport->NumberOfFunctions;
	ulNumberOfNames = pSelfExport->NumberOfNames + pGameExport->NumberOfNames;
	
	pulFunctions = (uint32 *) calloc (ulNumberOfFunctions * sizeof (*pulFunctions), 1);
	pSort = (sortnames_t *) calloc (ulNumberOfNames * sizeof (*pSort), 1);
	
	for (ulFunctionCount = 0; ulFunctionCount < pSelfExport->NumberOfFunctions; ulFunctionCount++)
		pulFunctions[ulFunctionCount] = RVA_TO_VA (g_hSelfLib, ((uint32 *) RVA_TO_VA (g_hSelfLib, pSelfExport->AddressOfFunctions))[ulFunctionCount]);
	
   for (ulNameCount = 0; ulNameCount < pSelfExport->NumberOfNames; ulNameCount++)
	{
		pSort[ulNameCount].ulName = RVA_TO_VA (g_hSelfLib, ((uint32 *) RVA_TO_VA (g_hSelfLib, pSelfExport->AddressOfNames))[ulNameCount]);
		pSort[ulNameCount].usNameOrdinal = ((uint16 *)RVA_TO_VA (g_hSelfLib, pSelfExport->AddressOfNameOrdinals))[ulNameCount];
	}
	
	for (i = 0; i < pGameExport->NumberOfFunctions; i++)
		pulFunctions[ulFunctionCount + i] = RVA_TO_VA (g_hGameLib, ((uint32 *) RVA_TO_VA (g_hGameLib, pGameExport->AddressOfFunctions))[i]);
	
   for (i = 0, ulListFix = 0; i < pGameExport->NumberOfNames; i++)
	{
		const char *szName = reinterpret_cast <const char *> (RVA_TO_VA (g_hGameLib, ((uint32 *) RVA_TO_VA (g_hGameLib, pGameExport->AddressOfNames))[i]));
		for (u = 0; u < ulNameCount; u++)
		{
			if (stricmp (szName, reinterpret_cast <const char *> (pSort[u].ulName)) == 0)
			{
				ulListFix -= 1;
				break;
			}
		}
		if(u < ulNameCount)
			continue;
		
		pSort[ulNameCount + i + ulListFix].ulName = reinterpret_cast <uint32> (szName);
		pSort[ulNameCount + i + ulListFix].usNameOrdinal = static_cast <uint16> (ulFunctionCount + (reinterpret_cast <uint16 *> (RVA_TO_VA (g_hGameLib, pGameExport->AddressOfNameOrdinals)))[i]);
   }
	
	ulNumberOfNames = ulNameCount + i + ulListFix;
	qsort (pSort, ulNumberOfNames, sizeof (*pSort), reinterpret_cast <int (*) (const void *, const void *)> (&SortOrdinalList));
	
	pulNames = static_cast <uint32 *> (VirtualAlloc (0, ulNumberOfNames * sizeof (*pulNames), MEM_COMMIT | MEM_TOP_DOWN, PAGE_READWRITE));
	pulNameOrdinals = static_cast <uint16 *> (VirtualAlloc (0, ulNumberOfNames * sizeof (*pulNameOrdinals), MEM_COMMIT | MEM_TOP_DOWN, PAGE_READWRITE));
	
	for (i = 0; i < ulNumberOfNames; i++)
	{
		pulNames[i] = pSort[i].ulName;
		pulNameOrdinals[i] = pSort[i].usNameOrdinal;
	}
	delete [] (pSort);
	
	for (i = 0; i < ulNumberOfFunctions; i++)
		pulFunctions[i] = VA_TO_RVA (g_hSelfLib, pulFunctions[i]);

	for (i = 0; i < ulNumberOfNames; i++)
	{
		pulNames[i] = VA_TO_RVA (g_hSelfLib, pulNames[i]);
		pulNameOrdinals[i] = static_cast <uint16> (VA_TO_RVA (g_hSelfLib, pulNameOrdinals[i]));
	}
	
	DWORD dwOldProtect = 0;
	if (!VirtualProtect (pSelfExport, sizeof (*pSelfExport), PAGE_READWRITE, &dwOldProtect))
      other_gFunctionTable.pfnSys_Error ("Couldn't initialize dynamic linkents. MemProtect Failed.");
	
	pSelfExport->Base = 1;
	pSelfExport->NumberOfFunctions = ulNumberOfFunctions;
	pSelfExport->NumberOfNames = ulNumberOfNames;
	pSelfExport->AddressOfFunctions = VA_TO_RVA (g_hSelfLib, pulFunctions);
	pSelfExport->AddressOfNames = VA_TO_RVA (g_hSelfLib, pulNames);
	pSelfExport->AddressOfNameOrdinals = VA_TO_RVA (g_hSelfLib, pulNameOrdinals);
	
	VirtualProtect (pSelfExport, sizeof (*pSelfExport), dwOldProtect, &dwOldProtect);
	
   return TRUE;
}


#elif defined (PF_LINUX32) || defined (PF_LINUX64)
   
// LINUX VERSION DOESN'T TESTED

DLSYM_FN g_pfnOriginalDlsym = NULL;
byte g_rgucDlsymBytes[BYTE_SIZE];
byte g_rgucDlsymOldBytes[BYTE_SIZE];

void *DlsymReplacement (void *hModule, const char *szFunctionName)
{
	static int iOriginalRestored = 0;
	int iWasOriginalRestored = iOriginalRestored;
   pthread_mutex_t kMutexDlsymReplacement = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
	
	pthread_mutex_lock (&kMutexDlsymReplacement);
	
	if (!iOriginalRestored)
	{
		memcpy ((void *) g_pfnOriginalDlsym, g_rgucDlsymOldBytes, BYTE_SIZE);
		iOriginalRestored = 1;
	}
		
	if((hModule != g_hSelfLib) || (g_hSelfLib == NULL) || (g_hGameLib == NULL))
	{
		void *pReturn = g_pfnOriginalDlsym (hModule, szFunctionName);
		
		if (g_hSelfLib && g_hGameLib)
		{
			if (!iWasOriginalRestored)
			{
				memcpy ((void *) g_pfnOriginalDlsym, g_rgucDlsymBytes, BYTE_SIZE);
				iOriginalRestored = 0;
			}
		}
		pthread_mutex_unlock (&kMutexDlsymReplacement);
		return pReturn;
	}
	void *pFunction = g_pfnOriginalDlsym (hModule, szFunctionName);
	
	if(pFunction == NULL)
		pFunction = g_pfnOriginalDlsym (g_hGameLib, szFunctionName);
	
	if(!iWasOriginalRestored)
	{
		memcpy ((void *) g_pfnOriginalDlsym, g_rgucDlsymBytes, BYTE_SIZE);
		iOriginalRestored = 0;
	}
	pthread_mutex_unlock(&kMutexDlsymReplacement);
	return pFunction;
}


int CombineExportTables (void)
{	
	g_pfnOriginalDlsym = (DLSYM_FN) dlsym (RTLD_DEFAULT, "dlsym");
	g_pfnOriginalDlsym (0, "dlopen");
	
	if (!(((byte *) g_pfnOriginalDlsym)[0] == 0xff || ((byte *) g_pfnOriginalDlsym)[1] == 0x25))
      other_gFunctionTable.pfnSys_Error ("Couldn't initialize dynamic linkents.");

#if defined (PF_LINUX64)
	byte *pucBytes = (byte *) ((void *) g_pfnOriginalDlsym);
	long lEndOfInstruction = (long) ((void *) g_pfnOriginalDlsym) + 6;
	
	pucBytes += 2;
   *(long *) &g_pfnOriginalDlsym = *(long *) (lEndOfInstruction + (long) *(int *) pucBytes);

#else

	*(long *) &g_pfnOriginalDlsym = (long) (** (void ***) ((char *)((void *) g_pfnOriginalDlsym) + 2));
#endif
   
   memcpy (g_rgucDlsymOldBytes, (void *) g_pfnOriginalDlsym, BYTE_SIZE);
	
#if defined (PF_LINUX64)
	((byte *) (void *) &g_rgucDlsymBytes[0])[0] = 0xff; 
	((byte *) (void *) &g_rgucDlsymBytes[0])[1] = 0x25; 
	((byte *) (void *) &g_rgucDlsymBytes[0])[2] = 0x0; 
	((byte *) (void *) &g_rgucDlsymBytes[0])[3] = 0x0; 
	((byte *) (void *) &g_rgucDlsymBytes[0])[4] = 0x0; 
	((byte *) (void *) &g_rgucDlsymBytes[0])[5] = 0x0; 

	void *pDblTarget = (void *)((void *) &DlsymReplacement); 
	memcpy(((char *)(void *) &g_rgucDlsymBytes[0] + 6), &pDblTarget, sizeof(pDblTarget)); 

#else

	((byte *) ((void *) &g_rgucDlsymBytes[0]))[0] = 0xff; 
	((byte *) ((void *) &g_rgucDlsymBytes[0]))[1] = 0x25; 
	
   void *pTarget = (void *) ((void *) &DlsymReplacement); 
	void **pDblTarget = &pTarget; 
	
   memcpy (((char *) ((void *) &g_rgucDlsymBytes[0]) + 2), &pDblTarget, sizeof (pDblTarget)); 
#endif

	long lStartOfPage = PAGE_ALIGN ((long) g_pfnOriginalDlsym) - PAGE_SIZE;
	long lSizeOfPages = 0;
	
	if((long) g_pfnOriginalDlsym + BYTE_SIZE > PAGE_ALIGN((long)g_pfnOriginalDlsym))
		lSizeOfPages = PAGE_SIZE * 2;
	else
		lSizeOfPages = PAGE_SIZE;
	
	if (mprotect ((void *) lStartOfPage, lSizeOfPages, PROT_READ | PROT_WRITE | PROT_EXEC))
      other_gFunctionTable.pfnSys_Error ("Couldn't initialize dynamic linkents. MemProtect Failed...");
	
	memcpy ((void *) g_pfnOriginalDlsym, g_rgucDlsymBytes, BYTE_SIZE);
	return TRUE;
}


#endif