//
// PODSteam - Loader for Old CS Bots
// Copyright (c) 2005, by Wei Mingzhi
// Copyright (c) 2006, by Dmitry Zhukov
//
// PodFix - POD-Bot 2.5 Bug Slayer
// Copyright (c) 2004, by Pierre-Marie Baty
//

#include "podsteam.h"

cvar_t pstm_podbuygrenades = {"pstm_podbuygrenades", "1", FCVAR_EXTDLL};
cvar_t pstm_podusescope = {"pstm_podusescope", "1", FCVAR_EXTDLL};

CPodFix::CPodFix (void)
{
   m_bBuyNades = TRUE;
   m_iUseScope = 1;
   m_fNadeBuyTime = 0;
   m_fRoundStartTime = 0;
}

void CPodFix::RegisterCVars (void)
{
   if (!g_bPODIngame) // no podbot, nothing to fix
      return;

   // ask the engine to register the CVARs this plugin uses
   CVAR_REGISTER (&pstm_podbuygrenades);
   CVAR_REGISTER (&pstm_podusescope);
}

void CPodFix::OnSpawnPost (edict_t *pent)
{
   if (!g_bPODIngame) // no podbot, nothing to fix
      return;

   // are we spawning a hostage ?
   if (strcmp (STRING (pent->v.classname), "hostage_entity") == 0)
   {
      pent->v.health = 0x0FFFFFFF; // max out the hostage life points
      pent->v.max_health = 0x0FFFFFFF;
   }
}

void CPodFix::OnFramePost (void)
{
   if (!g_bPODIngame) // no podbot, nothing to fix
      return;

   int client_index;
   edict_t *pClient;

   // retrieve our CVARs...
   m_bBuyNades = (CVAR_GET_FLOAT ("pstm_podbuygrenades") > 0);
   m_iUseScope = CVAR_GET_FLOAT ("pstm_podusescope");

   // is it time to make the bots buy grenades ?
   if ((m_fNadeBuyTime > 0) && (m_fNadeBuyTime < gpGlobals->time))
   {
      // cycle through all clients and find bots
      for (client_index = 1; client_index <= gpGlobals->maxClients; client_index++)
      {
         pClient = INDEXENT (client_index);

         if (!pClient || pClient->free || !(pClient->v.flags & FL_FAKECLIENT) || IsYaPB (pClient))
            continue; // skip invalid clients and real players

         CLIENT_COMMAND (pClient, "buyitem\n"); // make the bot enter the buy item menu
         CLIENT_COMMAND (pClient, "menuselect 4\n"); // make the bot buy a grenade
      }

      m_fNadeBuyTime = 0; // reset the timer so we won't do it twice
   }
}

void CPodFix::OnFindEntByStr (edict_t *pEdictStartSearchAfter, const char *pszField, const char *pszValue)
{
   if (!g_bPODIngame) // no podbot, nothing to fix
      return;

   // if we're resetting the map and we want the bots to buy more grenades...
   if (strcmp (pszValue, "info_map_parameters") == 0)
   {
      m_fRoundStartTime = gpGlobals->time + CVAR_GET_FLOAT ("mp_freezetime"); // compute round start date

      if (m_bBuyNades)
         m_fNadeBuyTime = gpGlobals->time + 3.0; // make them buy a grenade in 4 seconds
      else
         m_fNadeBuyTime = 0;
   }
}

bool CPodFix::OnRunPlayerMove (edict_t *fakeclient, const float *viewangles, float forwardmove, float sidemove, float upmove, unsigned short buttons, byte impulse, byte msec)
{
   if (!g_bPODIngame) // no podbot, nothing to fix
      return true;

   char bot_weaponmodel[256];
   bool has_bomb, has_zoomable_rifle, has_scopable_rifle, has_grenade, has_knife;
   Vector v_potential_enemy = Vector (0, 0, 0);

   // retrieve the weapon the bot is holding in its hands
   strcpy (bot_weaponmodel, STRING (fakeclient->v.weaponmodel));

   // determine if it is holding a bomb, or any sort of scopable rifle, a grenade or a knife
   has_bomb = (strcmp (bot_weaponmodel, "models/p_c4.mdl") == 0);
   has_zoomable_rifle = (!has_bomb
                         && ((strcmp (bot_weaponmodel, "models/p_aug.mdl") == 0)
                             || (strcmp (bot_weaponmodel, "models/p_sg552.mdl") == 0)));
   has_scopable_rifle = (!has_bomb && !has_zoomable_rifle
                         && ((strcmp (bot_weaponmodel, "models/p_awp.mdl") == 0)
                             || (strcmp (bot_weaponmodel, "models/p_g3sg1.mdl") == 0)
                             || (strcmp (bot_weaponmodel, "models/p_scout.mdl") == 0)
                             || (strcmp (bot_weaponmodel, "models/p_sg550.mdl") == 0)));
   has_grenade = (!has_bomb && !has_zoomable_rifle && !has_scopable_rifle
                  && ((strcmp (bot_weaponmodel, "models/p_flashbang.mdl") == 0)
                      || (strcmp (bot_weaponmodel, "models/p_he.mdl") == 0)
                      || (strcmp (bot_weaponmodel, "models/p_smokegrenade.mdl") == 0)));
   has_knife = (!has_bomb && !has_zoomable_rifle && !has_scopable_rifle && !has_grenade
                && (strcmp (bot_weaponmodel, "models/p_knife.mdl") == 0));

   v_potential_enemy = FakeclientSeesEnemyPlayer (fakeclient);

   // is this bot about to plant the bomb ?
   if (has_bomb)
   {
      fakeclient->v.button = (fakeclient->v.button | IN_ATTACK) & ~IN_JUMP; // force it to plant the bomb right here right now
      (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
      return false;
   }

   // else is this bot holding a sniper rifle AND we take over the sniping code ?
   if ((has_zoomable_rifle || has_scopable_rifle) && (m_iUseScope > 0))
   {
      // is this bot trying to lock onto an enemy ?
      if (v_potential_enemy != Vector (0, 0, 0))
      {
         // is it a scoped weapon (sniper rifle) ?
         if (has_scopable_rifle)
         {
            // if the enemy is very far and bot is allowed to scope full zoom and bot is not scoping yet...
            if (((v_potential_enemy - fakeclient->v.origin).Length () > 600) && (m_iUseScope > 1) && (fakeclient->v.fov > FOV_SCOPED_FAR))
            {
               if (RANDOM_LONG (0, 100) > 50)
                  fakeclient->v.button = (fakeclient->v.button | IN_ATTACK2) & ~IN_ATTACK; // scope until the highest zoom mode
               else
                  fakeclient->v.button &= IN_ATTACK2;
               (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
               return false;
            }

            // else if it's not immediately nearby...
            else if (((v_potential_enemy - fakeclient->v.origin).Length () > 150) && (fakeclient->v.fov > FOV_SCOPED_SHORT))
            {
               if (RANDOM_LONG (0, 100) > 50)
                  fakeclient->v.button = (fakeclient->v.button | IN_ATTACK2) & ~IN_ATTACK; // scope to the first zoom mode (because scoped weapons have no crosshair, and bot NEEDS a crosshair :))
               else
                  fakeclient->v.button &= IN_ATTACK2;
               (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
               return false;
            }
         }

         // else it is a zoomed weapon (Steyr Aug, Sig SG-552)
         else if (((v_potential_enemy - fakeclient->v.origin).Length () > 400) && (fakeclient->v.fov > FOV_ZOOMED))
         {
            if (RANDOM_LONG (0, 100) > 50)
               fakeclient->v.button = (fakeclient->v.button | IN_ATTACK2) & ~IN_ATTACK; // use the zoom when appropriate
            else
               fakeclient->v.button &= IN_ATTACK2;
            (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
            return false;
         }
      }

      // else if the bot seems not to try to lock onto anyone, reset the weapon zoom to normal
      else if (fakeclient->v.fov < FOV_NORMAL)
      {
         if (RANDOM_LONG (0, 100) > 50)
            fakeclient->v.button = (fakeclient->v.button | IN_ATTACK2) & ~IN_ATTACK; // force it to reset the zoom right now so that the bot sees where it put its feet, at least!
         else
            fakeclient->v.button &= IN_ATTACK2;
         (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
         return false;
      }

      // when everything is right, don't let the bot code handle the scope
      fakeclient->v.button &= ~IN_ATTACK2; // prevent POD-bot from zooming or unzooming on its own
      (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
      return false;
   }

   // else if it's holding a grenade and the enemy is in sight, there's an emergency to throw it !
   else if (has_grenade && (v_potential_enemy != Vector (0, 0, 0)))
   {
      if (RANDOM_LONG (0, 100) > 50)
         fakeclient->v.button |= IN_ATTACK; // force it to throw the grenade right now
      else
         fakeclient->v.button &= ~IN_ATTACK;
      (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
      return false;
   }

   // else if it's holding a knife, make it stab occasionally (more realistic ? also helps fixing stuck to enemy bug)
   else if (has_knife && (v_potential_enemy != Vector (0, 0, 0)))
   {
      if (RANDOM_LONG (0, 100) > 98)
         fakeclient->v.button |= IN_ATTACK; // force it to stab right now
      (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
      return false;
   }

   // additionnally, if the bot is still in freeze time, it's no use jumping nor ducking :)
   if (m_fRoundStartTime > gpGlobals->time)
   {
      fakeclient->v.button &= ~(IN_JUMP | IN_DUCK); // prevent the bot from jumping like crazy
      (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
      return false;
   }

   return true;
}

bool CPodFix::OnTraceLine (const float *v1, const float *v2, int fNoMonsters, edict_t *pentToSkip, TraceResult *ptr)
{
   if (!g_bPODIngame) // no podbot, nothing to fix
      return true;

   TRACE_LINE (v1, v2, fNoMonsters, pentToSkip, ptr);

   // check to see if a hostage was hit	
   if (strcmp (STRING (ptr->pHit->v.classname), "hostage_entity") == 0 )
   {
      ptr->flFraction = 0; // don't even reach the hostage, disallowing damage for it
      ptr->pHit->v.health = 0x0FFFFFFF; // max out the hostage life points again
      ptr->pHit->v.max_health = 0x0FFFFFFF;
      return false;
   }
   return true;
}

Vector CPodFix::FakeclientSeesEnemyPlayer (edict_t *pFakeclient)
{
   int client_index;
   edict_t *pPlayer = NULL;
   TraceResult tr;

   if (FNullEnt (pFakeclient))
      return (Vector (0, 0, 0)); // reliability check

   // cycle through all players
   for (client_index = 1; client_index < gpGlobals->maxClients; client_index++)
   {
      pPlayer = INDEXENT (client_index);

      if (FNullEnt (pPlayer) || pPlayer->free || (pPlayer == pFakeclient)
          || (pPlayer->v.effects & EF_NODRAW) || (GetTeam (pFakeclient) == GetTeam (pPlayer)))
         continue; // skip invalid players, invisible ones, ourselves and teammates

      MAKE_VECTORS (pFakeclient->v.v_angle);
      if (AngleOfVectors (gpGlobals->v_forward, pPlayer->v.origin - pFakeclient->v.origin) > pFakeclient->v.fov / 2)
         continue; // skip players outside bot's field of view

      MAKE_VECTORS (VecToAngles (pPlayer->v.origin - pFakeclient->v.origin));

      // trace a line from bot's eye to this player's head
      TRACE_LINE (pFakeclient->v.origin + pFakeclient->v.view_ofs, pPlayer->v.origin + pPlayer->v.view_ofs, ignore_monsters, pFakeclient, &tr);
      if (tr.flFraction == 1.0)
         return (tr.vecEndPos); // enemy in sight if the line traced all the way up to its destination

      // trace a line from bot's eye to this player's body
      TRACE_LINE (pFakeclient->v.origin + pFakeclient->v.view_ofs, pPlayer->v.origin, ignore_monsters, pFakeclient, &tr);
      if (tr.flFraction == 1.0)
         return (tr.vecEndPos); // enemy in sight if the line traced all the way up to its destination

      // trace a line from bot's eye to this player's left arm
      TRACE_LINE (pFakeclient->v.origin + pFakeclient->v.view_ofs, pPlayer->v.origin - (gpGlobals->v_right * 16), ignore_monsters, pFakeclient, &tr);
      if (tr.flFraction == 1.0)
         return (tr.vecEndPos); // enemy in sight if the line traced all the way up to its destination

      // trace a line from bot's eye to this player's right arm
      TRACE_LINE (pFakeclient->v.origin + pFakeclient->v.view_ofs, pPlayer->v.origin + (gpGlobals->v_right * 16), ignore_monsters, pFakeclient, &tr);
      if (tr.flFraction == 1.0)
         return (tr.vecEndPos); // enemy in sight if the line traced all the way up to its destination
   }

   return (Vector (0, 0, 0)); // looks like this bot has no enemy in sight
}

int GetTeam (edict_t *pEntity)
{
   short t = *(short *)g_engfuncs.pfnInfoKeyValue (g_engfuncs.pfnGetInfoKeyBuffer (pEntity), "model");

   if ((t == (('e' << 8) + 't')) || // TError
       (t == (('e' << 8) + 'l')) || // LEet
       (t == (('r' << 8) + 'a')) || // ARctic
       (t == (('u' << 8) + 'g')) || // GUerilla
       (t == (('i' << 8) + 'm')))   // MIlitia
      return TEAM_TERRORIST;

   return TEAM_COUNTER_TERRORIST; // URban, GSg9, SAs, GIgn, VIp, SPetsnaz
}
