//
// PODSteam - Loader for Old CS Bots
// Copyright (c) 2005, by Wei Mingzhi
// Copyright (c) 2006, by Dmitry Zhukov
//

#ifndef PODSTEAM_H
#define PODSTEAM_H

#include "extdll.h"
#include "stdio.h" 

#pragma warning (disable: 4702) // unreachable code
#pragma warning (disable: 4312) // pointer truncation from 'dynlib_t' to 'uint32'
#pragma warning (disable: 4311) // conversion from 'const uint32' to 'const char *' of greater size
#pragma warning (disable: 4267) // conversion from 'size_t' to 'int', possible loss of data
#pragma warning (disable: 4996) // visual c++ secure notes about unsafe deprecated CRT function calls

// detects the build platform
#if defined (__linux__) || defined (__debian__) || defined (__linux)
	#define PF_LINUX32 1
#elif defined (__x86_64__) || defined (__amd64__)
   #define PF_LINUX64 1
#elif defined (_WIN32)
	#define PF_WIN32 1
#endif

// detects the compiler
#if defined (_MSC_VER)
   #define CT_MSVC _MSC_VER
#elif defined (__BORLANDC__)
   #define CT_BORLAND __BORLANDC__
#endif

// configure export macros
#if defined (CT_MSVC)
   #define export extern "C" __declspec (dllexport)
#elif defined (PF_LINUX32) || defined (PF_LINUX64) || defined (CT_BORLAND)
   #define export extern "C"
#endif

#include <util.h>
#include <enginecallback.h>

// operating system specific macros, functions and typedefs
#ifdef PF_WIN32

   #include <direct.h>
   #include <io.h>

   #define DLL_ENTRYPOINT int STDCALL DllMain (HINSTANCE hModule, DWORD dwReason, LPVOID)
   #define DLL_DETACHING (dwReason == DLL_PROCESS_DETACH)
   #define DLL_RETENTRY return TRUE
   #define DLL_GIVEFNPTRSTODLL extern "C" void STDCALL

   // are we under the microsoft visual c compiler
   #if defined (CT_MSVC) && (CT_MSVC > 1000) 
      // specify export parameter (c) Jozef Wagner
      #pragma comment (linker,"/EXPORT:GiveFnptrsToDll=_GiveFnptrsToDll@8,@1")
      #pragma comment (linker,"/SECTION:.data,RW")
   #endif

   // dynamic linkents
   #define RVA_TO_VA(kBase, kRva) ((uint32) kBase + (uint32) kRva)
   #define VA_TO_RVA(kBase, kVa) ((uint32) kVa - (uint32) kBase)

   typedef HINSTANCE dynlib_t;
   typedef FARPROC dynsym_t;

   typedef int (FAR *GETENTITYAPI) (DLL_FUNCTIONS *, int);
   typedef void (FAR *ENTITY_FN) (entvars_t *);
   typedef void (STDCALL *GIVEFNPTRSTODLL) (enginefuncs_t *, globalvars_t *);

   inline dynlib_t OpenLibrary (const char *cszFilename) { return LoadLibrary (cszFilename); }
   inline dynsym_t GetProcSymbol (dynlib_t hLib, const char *cszSymbol) { return GetProcAddress (hLib, cszSymbol); }
   inline int CloseLibrary (dynlib_t hLib) { return FreeLibrary (hLib); }
   inline dynlib_t HandleFromMemPtr (void *pMem) { MEMORY_BASIC_INFORMATION kMbi; if (!VirtualQuery (pMem, &kMbi, sizeof (kMbi)) || (kMbi.State != MEM_COMMIT) || (!kMbi.AllocationBase)) return NULL; return (dynlib_t) kMbi.AllocationBase; }

#elif defined (PF_LINUX32) || defined (PF_LINUX64)

   #include <unistd.h>
   #include <dlfcn.h>
   #include <errno.h>
   
   #include <asm/page.h>
   #include <sys/stat.h>
   #include <sys/mman.h>
   #include <pthread.h>

   #define DLL_ENTRYPOINT void _fini (void)
   #define DLL_DETACHING TRUE
   #define DLL_RETENTRY return
   #define DLL_GIVEFNPTRSTODLL extern "C" void

   // dynamic linkents
   #define PAGE_ALIGN(kAddr) (((kAddr) + PAGE_SIZE - 1) & PAGE_MASK)

   #if defined (PF_LINUX64)
      #define BYTE_SIZE (6 + sizeof (void *))
   #else
      #define BYTE_SIZE (2 + sizeof (void *))
   #endif

   typedef void *dynlib_t;
   typedef void *dynsym_t;

   typedef int (*GETENTITYAPI) (DLL_FUNCTIONS *, int);
   typedef void (*ENTITY_FN) (entvars_t *);
   typedef void (*GIVEFNPTRSTODLL) (enginefuncs_t *, globalvars_t *);
   typedef void *(*DLSYM_FN)(dynlib_t, const char *);

   inline dynlib_t OpenLibrary (const char *cszFilename) { return dlopen (cszFilename, RTLD_NOW); }
   inline dynsym_t GetProcSymbol (dynlib_t hLib, const char *cszSymbol) { return dlsym (hLib, cszSymbol); }
   inline int CloseLibrary (dynlib_t hLib) { return dlclose (hLib); }
   inline dynlib_t HandleFromMemPtr (void *pMem) { Dl_info kDli; memset (&kDli, 0, sizeof (kDli)); if (dladdr (pMem, &kDli)) return (dlopen (kDli.dli_fname, RTLD_NOW)); else return NULL; }

#else
   #error "Platform unrecognized."

#endif      

#include <assert.h>

// stuff for function sort in win32
typedef struct 
{
	uint32 ulName;
	uint16 usNameOrdinal;
} sortnames_t;

// BOTAIM PLUGIN
class CBotAim
{
private:
   int         iTurnMethod;
   bool        m_bAimFix;
   int         m_iTurnSkill;
   Vector      m_vViewAngle[32];

public:
   void        RegisterCVars (void);
   void        PointGun (edict_t *pEdict, bool save_angles);
   bool        IsAim (edict_t *pEdict);

   void        OnFramePost (void);
   bool        OnTraceLine (const float *v1, const float *v2, int fNoMonsters, edict_t *pentToSkip, TraceResult *ptr);
};

// PODFIX PLUGIN
class CPodFix
{
private:
   bool        m_bBuyNades;
   int         m_iUseScope;
   float       m_fNadeBuyTime;
   float       m_fRoundStartTime;

public:
   CPodFix     (void);
   ~CPodFix    (void) { };

   void        RegisterCVars (void);
   
   void        OnSpawnPost (edict_t *pent);
   void        OnFramePost (void);

   void        OnFindEntByStr (edict_t *pEdictStartSearchAfter, const char *pszField, const char *pszValue);
   bool        OnRunPlayerMove (edict_t *fakeclient, const float *viewangles, float forwardmove, float sidemove, float upmove, unsigned short buttons, byte impulse, byte msec);
   bool        OnTraceLine (const float *v1, const float *v2, int fNoMonsters, edict_t *pentToSkip, TraceResult *ptr);

   Vector      FakeclientSeesEnemyPlayer (edict_t *pFakeclient);
};

enum
{
   Menu_None = 0,
   Menu_Buy,
   Menu_BuyPistol,
   Menu_BuyRifle,
   Menu_BuyMachineGun,
   Menu_BuyShotgun,
   Menu_BuySubMachineGun,
   Menu_BuyItem,
};

struct tShowMenu
{
   edict_t *pent;
   signed int iShort;
   signed int iChar;
   signed int iByte;
   char szActualMessage[32];
};

// turn methods
#define AIM_GENUINE 0
#define AIM_RACC_OLD 1
#define AIM_RACC 2

// Counter-Strike team IDs
#define TEAM_TERRORIST 1
#define TEAM_COUNTER_TERRORIST 2

// normal / zoomed / scoped FOV widths
#define FOV_NORMAL 90
#define FOV_ZOOMED 55
#define FOV_SCOPED_SHORT 40
#define FOV_SCOPED_FAR 15 // AWP is the exception, it can scope down to 10 degrees

extern dynlib_t g_hGameLib;
extern dynlib_t g_hSelfLib;
extern DLL_FUNCTIONS other_gFunctionTable;
extern bool g_bPODIngame;

int CombineExportTables (void);
int GetTeam (edict_t *pEntity);
float AngleOfVectors (Vector v1, Vector v2);
float WrapAngle (float angle);
Vector VecToAngles (const Vector &v_vector);
Vector WrapAngles (Vector angles);

// yapb doesn't set "ah" (since it's changed to _ah in cs1.5), in the infokey buffer so skip it because there is nothing to fix or correct
// used when running old & new bot at the same time
inline bool IsYaPB (edict_t *pent)
{
   char *infobuffer = (*g_engfuncs.pfnGetInfoKeyBuffer) (pent);

   if (!(*g_engfuncs.pfnInfoKeyValue) (infobuffer, "ah")[0])
      return true;
   return false;
}

#include <vector>

#endif
